# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Configuration definition objects."""

__all__ = ['Configuration']

from .base import CompositeModelElement
from .cat import Category

from ..parser.resolver.core import (
    DEFAULT_SAFE, DEFAULT_BESTEFFORT, DEFAULT_SCOPE
)


class Configuration(CompositeModelElement):
    """Manage conf such as a list of Categories.

    The order of categories permit to ensure param overriding.
    """

    __contenttype__ = Category  #: content type.

    __slots__ = (
        'safe', 'besteffort', 'scope'
    ) + CompositeModelElement.__slots__

    def __init__(
            self, safe=DEFAULT_SAFE, besteffort=DEFAULT_BESTEFFORT,
            scope=DEFAULT_SCOPE, *args, **kwargs
    ):

        super(Configuration, self).__init__(*args, **kwargs)

        self.safe = safe
        self.besteffort = besteffort
        self.scope = scope

    def resolve(
            self, configurable=None, scope=None, safe=None, besteffort=None
    ):
        """Resolve all parameters.

        :param Configurable configurable: configurable to use for foreign
            parameter resolution.
        :param dict scope: variables to use for parameter expression evaluation.
        :param bool safe: safe execution (remove builtins functions).
        :raises: Parameter.Error for any raised exception.
        """

        if scope is None:
            scope = self.scope

        if safe is None:
            safe = self.safe

        if besteffort is None:
            besteffort = self.besteffort

        for category in self.values():

            for param in category.values():

                param.resolve(
                    configurable=configurable, conf=self,
                    scope=scope, safe=safe, besteffort=besteffort
                )

    def param(self, pname, cname=None, history=0):
        """Get parameter from a category and history.

        :param str pname: parameter name.
        :param str cname: category name. Default is the last registered.
        :param int history: historical param value from specific category or
            final parameter value if cname is not given. For example, if history
            equals 1 and cname is None, result is the value defined just before
            the last parameter value if exist. If cname is given, the result
            is the parameter value defined before the category cname.
        :rtype: Parameter
        :raises: NameError if pname or cname do not exist."""

        result = None

        category = None

        categories = []  # list of categories containing input parameter name

        for cat in self.values():

            if pname in cat:

                categories.append(cat)

                if cname == cat.name:
                    break

        if cname is not None and (
                not categories or categories[-1].name != cname
        ):
            raise NameError('Category {0} does not exist.'.format(cname))

        categories = categories[:max(1, len(categories) - history)]

        for category in categories:
            if pname in category:
                if result is None:
                    result = category[pname].copy()

                else:
                    result.update(category[pname])

        return result


def configuration(*cats):
    """Quick instanciaton of Configuration with categories."""

    return Configuration(melts=cats)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------


from unittest import main

from b3j0f.utils.ut import UTCase

from ..core import Configurable, applyconfiguration
from ...model.conf import configuration
from ...model.cat import category
from ...model.param import Parameter

from ...driver.test.base import TestConfDriver

from tempfile import NamedTemporaryFile

from six import add_metaclass


class ConfigurableTest(UTCase):

    def setUp(self):

        self.paths = (
            NamedTemporaryFile().name,
            NamedTemporaryFile().name
        )

        self.configurable = Configurable()

        self.conf = configuration(
            category(
                'A',
                Parameter('a', value='a', ptype=str),
                Parameter('_', value=2, ptype=int),
                Parameter('error', ptype=float, svalue='error')
            ),
            category(
                'B',
                Parameter('a', value='b', ptype=str),
                Parameter('b', value='b', ptype=str)
            )
        )

    def test_oldstyle(self):

        @Configurable()
        class TestA:
            pass

        TestA()

        @Configurable()
        class TestB:
            def __init__(self):
                pass

        TestB()

        @Configurable()
        class TestC(TestA):
            def __init__(self):
                TestA.__init__(self)

        TestC()

    def test_configuration_files(self):

        configurable = Configurable(paths=self.paths)

        self.assertEqual(configurable.paths, self.paths)

        configurable = Configurable()
        configurable.paths = self.paths

        self.assertEqual(configurable.paths, self.paths)

    def test_configure_to_reconfigure_param(self):
        """Test to reconfigure an object with targets parameter."""

        class ToConfigure(object):
            """Class to configure."""

            def __init__(self):
                super(ToConfigure, self).__init__()
                self.test = None

        target = ToConfigure()

        param = 'test'

        conf = configuration(category('TEST', Parameter(param, value=True)))

        self.configurable.configure(conf=conf, targets=[target])
        self.assertTrue(target.test)

    def test_configure_without_inheritance(self):
        """Test to configure an object without inheritance."""

        @Configurable(conf=category('TEST', Parameter('test', value=True)))
        class BaseTest(object):
            """base Class to configure."""

        class Test(BaseTest):
            """Class to configure."""

        targets = Test()

        self.assertTrue(targets.test)

    def test_parser_inheritance(self):

        @Configurable()
        class BaseTest(object):
            pass

        @Configurable()
        @Configurable()
        class SubTest(BaseTest):
            pass

        class Test(SubTest):
            pass

        configurables = Configurable.get_annotations(BaseTest)

        self.assertEqual(len(configurables), 1)

        configurables = Configurable.get_annotations(SubTest)

        self.assertEqual(len(configurables), 3)

        configurables = Configurable.get_annotations(Test)

        self.assertEqual(len(configurables), 3)

    def test_class_callparams(self):
        """Test to call params on a class."""

        @Configurable(
            conf=[
                Parameter('test0', value=True),
                Parameter('test1', value=False)
            ]
        )
        class Test(object):

            def __init__(self, test0=None):

                super(Test, self).__init__()

                self.test0 = test0

        test = Test()

        self.assertTrue(test.test0)
        self.assertFalse(test.test1)

    def test_metaclass(self):
        """Test to configurate a metaclass."""

        @Configurable(conf=Parameter('test', value=True))
        class MetaTest(type):
            pass

        @add_metaclass(MetaTest)
        class Test(object):
            pass

        test = Test()

        self.assertTrue(test.test)

    def test_function_callparams(self):
        """Test to call params on a function."""

        @Configurable(conf=category('', Parameter('test', value=True)))
        def twist(test=None):
            return test

        value = twist()

        self.assertTrue(value)

        value = twist(False)

        self.assertFalse(value)

        Configurable.get_annotations(twist)[0].conf['']['test'].value = 1

        value = twist()

        self.assertEqual(value, 1)

    def test_inheritance(self):
        """Test with inheritance between targets objects."""

        conf0 = configuration(
            category(
                'test0',
                Parameter('test0', value=0)
            ),
            category(
                'test1',
                Parameter('test1', value=1)
            )
        )

        conf1 = configuration(
            category(
                'test1',
                Parameter('test1', value=2)
            )
        )

        @Configurable(conf=conf0)
        class Parent(object):
            pass

        self.assertEqual(Parent().test0, 0)
        self.assertEqual(Parent().test1, 1)

        @Configurable(conf=conf1)
        class Child(Parent):
            pass

        self.assertEqual(Child().test0, 0)
        self.assertEqual(Child().test1, 2)

    def test_applyconfiguration(self):
        """Test the function applyconfiguration."""

        conf = configuration(category('', Parameter('test', value=True)))

        @Configurable(conf=conf)
        class Test(object):
            pass

        test = Test()

        self.assertTrue(test.test)

        test.test = False

        applyconfiguration(targets=[test])

        self.assertTrue(test.test)

        class Test(object):
            pass

        test = Test()

        self.assertFalse(hasattr(test, 'test'))

        applyconfiguration(targets=[test], conf=conf)

        self.assertTrue(test.test)

    def test_annotation(self):
        """Test to use a configurable instancec such as a decodator."""

        configurable = Configurable(
            conf=configuration(category('', Parameter('test', value=True)))
        )

        @configurable
        class Test(object):
            pass

        test = Test()

        self.assertTrue(test.test)

        test.test = False

        applyconfiguration(targets=[test])

        self.assertTrue(test.test)

    def test_object(self):
        """Test to configure objects."""

        configurable = Configurable(
            conf=configuration(category('', Parameter('test', value=True)))
        )

        class Test(object):
            pass

        test = Test()

        configurable(test)
        configurable.applyconfiguration(targets=[test])

        self.assertTrue(test.test)

        test.test = False

        applyconfiguration(targets=[test])

        self.assertTrue(test.test)

    def test_multiconf(self):
        """Test to get configuration from multi resources."""

        tcd0, tcd1 = TestConfDriver(), TestConfDriver()

        tcd0.confbypath['test0'] = configuration(
            category(
                'test',
                Parameter('test0', value='0')
            )
        )
        tcd0.confbypath['test1'] = configuration(
            category(
                'test',
                Parameter('test1', value='1')
            )
        )
        tcd1.confbypath['test1'] = configuration(
            category(
                'test',
                Parameter('test2', svalue='=@test0/test.test0 + @test1/test.test1'),
                Parameter('test3', value=3)
            )
        )

        configurable = Configurable(drivers=[tcd0, tcd1])

        configurable.applyconfiguration(
            targets=[configurable], paths=['test0', 'test1']
        )

        self.assertEqual(configurable.test0, '0')
        self.assertEqual(configurable.test1, '1')
        self.assertEqual(configurable.test2, '01')
        self.assertEqual(configurable.test3, 3)

    def test_safe(self):
        """Test to configurate a configurable in a safe context."""

        conf = configuration(
            category(
                'test',
                Parameter('test', svalue='=open')
            )
        )

        configurable = Configurable(conf=conf, autoconf=False)

        self.assertRaises(
            Parameter.Error,
            configurable.applyconfiguration,
            targets=configurable, paths='test'
        )

    def test_unsafe(self):
        """Test to configurate a configurable in an unsafe context."""

        conf = configuration(
            category(
                'test',
                Parameter('test', svalue='=int')
            )
        )

        configurable = Configurable(conf=conf, safe=False)

        configurable.applyconfiguration(targets=[configurable], paths='test')

        self.assertIs(configurable.test, int)

    def test_subconf(self):
        """Test sub configuration."""

        class Test(object):
            pass

        class Test2(object):
            def __init__(self, param=None):
                self.param = param

        conf = configuration(
            category(
                'test',
                Parameter('subtest', value=Test),
                Parameter('subtest2', value=Test2)
            ),
            category(
                ':subtest',
                Parameter('param', svalue='=@subtest'),
                Parameter('attr', value=True)
            ),
            category(
                ':subtest:param',
                Parameter('param', svalue='=@subtest'),
                Parameter('attr', value=True)
            )
        )

        test = Test()

        applyconfiguration(conf=conf, targets=[test])

        self.assertIs(test.subtest2, Test2)
        self.assertIsInstance(test.subtest, Test)
        self.assertTrue(test.subtest.attr)
        self.assertIsInstance(test.subtest.param, Test)
        self.assertIs(test.subtest.param.param, Test)
        self.assertTrue(test.subtest.param.attr)

        oldsubtestparam = test.subtest.param

        conf[':subtest:param']['attr'].value = False

        applyconfiguration(conf=conf, targets=[test], keepstate=True)

        self.assertIs(test.subtest2, Test2)
        self.assertIsInstance(test.subtest, Test)
        self.assertTrue(test.subtest.attr)
        self.assertIsInstance(test.subtest.param, Test)
        self.assertIs(test.subtest.param.param, Test)
        self.assertFalse(test.subtest.param.attr)

        self.assertIs(test.subtest.param, oldsubtestparam)

        applyconfiguration(conf=conf, targets=[test], keepstate=False)

        self.assertIs(test.subtest2, Test2)
        self.assertIsInstance(test.subtest, Test)
        self.assertTrue(test.subtest.attr)
        self.assertIsInstance(test.subtest.param, Test)
        self.assertIs(test.subtest.param.param, Test)
        self.assertFalse(test.subtest.param.attr)

        self.assertIsNot(test.subtest.param, oldsubtestparam)

    def test_ptype(self):
        """Test application of ptype."""

        @Configurable(
            conf=[
                Parameter('test', ptype=int, svalue='1'),
                Parameter('ex', svalue='2', ptype=int)
            ]
        )
        class Test(object):

            def __init__(self, test=None, *args, **kwargs):

                super(Test, self).__init__(*args, **kwargs)

                self.testy = test

        test = Test()

        self.assertEqual(test.testy, 1)
        self.assertFalse(hasattr(test, 'test'))
        self.assertEqual(test.ex, 2)

        applyconfiguration(
            targets=[test], conf=[
                Parameter('test', svalue='2'),
                Parameter('ex', svalue='3')
            ]
        )

        self.assertEqual(test.testy, 1)
        self.assertEqual(test.test, 2)
        self.assertEqual(test.ex, 3)

        Configurable.get_annotations(test)[0].applyconfiguration(
            targets=[test], conf=[
                Parameter('test', svalue='3'),
                Parameter('ex', svalue='4', ptype=bool)
            ]
        )

        self.assertEqual(test.testy, 1)
        self.assertEqual(test.test, 3)
        self.assertTrue(test.ex)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Module of configuration drivers.

A configuration driver resolve a configuration path with conf resources paths.

A conf path is a configuration path understandable by several drivers.
A resource path is a configuration path understandable only by one type of
driver (and all its sub types of course...).

In order to implement your self drivers, you have to implement those methods:

- rscpaths(path): get resource paths from one configuration path.
- _getconf(rscpath, logger): get one configuration from one resource path.
- _setconf(rscpath, logger): put one configuration from one resource path.
"""

__all__ = ['ConfDriver']

from ..model.conf import Configuration
from ..model.cat import Category

from traceback import format_exc

from six import reraise


class ConfDriver(object):
    """Driver dedicated to get/set configuration from relative paths.

    A driver associate a configuration path to

    - a specific configuration resource.
    - specific resource paths.
    """

    class Error(Exception):
        """Handle conf driver errors."""

    def rscpaths(self, path):
        """Get resource paths related to input configuration path.

        :param str path: configuration path.
        :return: resource paths which could be used by the method setconf.
        :rtype: list
        """

        raise NotImplementedError()

    def resource(self):
        """Get a default and empty resource.

        :return: default specific resource.
        """

        raise NotImplementedError()

    def pathresource(self, rscpath=None, logger=None):
        """Returns specific resource.

        :param str rscpath: resource path.
        :param Logger logger: logger to use.
        :param bool error: raise internal error if True (False by default).
        :param bool force: create the resource even if rscpath does not exist.
        :return: specific configuration resource.
        """

        result = None

        try:
            result = self._pathresource(rscpath=rscpath)

        except Exception as ex:
            if logger is not None:
                msg = 'Error while getting resource from {0}.'.format(rscpath)
                full_msg = '{0} {1}: {2}'.format(msg, ex, format_exc())
                logger.error(full_msg)

        return result

    def getconf(self, path, conf=None, logger=None):
        """Parse a configuration path with input conf and returns
        parameters by param name.

        :param str path: conf resource path to parse and from get parameters.
        :param Configuration conf: conf to fill with path values and
            conf param names.
        :param Logger logger: logger to use in order to trace
            information/error.
        :rtype: Configuration
        """

        result = conf

        pathconf = None

        rscpaths = self.rscpaths(path=path)

        for rscpath in rscpaths:

            pathconf = self._getconf(rscpath=rscpath, logger=logger, conf=conf)

            if pathconf is not None:

                if result is None:
                    result = pathconf

                else:
                    result.update(pathconf)

        return result

    def setconf(self, conf, rscpath, logger=None):
        """Set input conf in input path.

        :param Configuration conf: conf to write to path.
        :param str rscpath: specific resource path to use.
        :param Logger logger: used to log info/errors.
        :param bool error: raise catched errors.
        :raises: ConfDriver.Error in case of error and input error.
        """

        resource = self.pathresource(rscpath=rscpath, logger=logger)

        if resource is None:
            resource = self.resource()

        try:
            self._setconf(conf=conf, resource=resource, rscpath=rscpath)

        except Exception as ex:
            if logger is not None:
                msg = 'Error while setting conf to {0}.'.format(rscpath)
                full_msg = '{0} {1}: {2}'.format(msg, ex, format_exc())
                logger.error(full_msg)
                reraise(self.Error, self.Error(msg))

    def _getconf(self, rscpath, logger=None, conf=None):
        """Get specific conf from one driver path.

        :param str rscpath: resource path.
        :param Logger logger: logger to use.
        """

        result = None

        resource = self.pathresource(rscpath=rscpath, logger=logger)

        if resource is not None:

            for cname in self._cnames(resource=resource):

                category = Category(name=cname)

                if result is None:
                    result = Configuration()

                result += category

                for param in self._params(resource=resource, cname=cname):

                    if conf is not None:
                        confparam = None

                        if cname in conf and param.name in conf[cname]:
                            confparam = conf[cname][param.name]

                        else:
                            confparam = conf.param(pname=param.name)

                        if confparam is not None:

                            param.update(confparam)

                    category += param

        return result

    def _setconf(self, conf, resource, rscpath):
        """Set input conf to input resource.

        :param Configuration conf: conf to write to path.
        :param resource: driver configuration resource.
        :param str rscpath: specific resource path to use.
        :param Logger logger: used to log info/errors.
        :raises: ConfDriver.Error if input error and an error has been raised.
        """

        raise NotImplementedError()

    def _pathresource(self, rscpath):
        """Method to override in order to get specific conf resource from a
        resource path.

        :param str rscpath: resource path.
        :param Logger logger: logger to use.
        :param bool error: if True (False by default), raise a ConfDriver Error.
        :param bool force: create resource even if rscpath does not exist.
        :return: specific configuration resource.
        """

        raise NotImplementedError()

    def _cnames(self, resource):
        """Get resource category names.

        :param resource: resource from where get category names.
        :param Logger logger: logger to use.
        :return: resource category names.
        :rtype: list
        """

        raise NotImplementedError()

    def _params(self, resource, cname):
        """Get list of category parameters.

        :param resource: resource from where get parameters.
        :param str cname: related category name.
        :param Logger logger: logger to use.
        :return: list of category parameters.
        :rtype: list
        """

        raise NotImplementedError()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

__all__ = ['FileConfDriver']

"""This module specifies file drivers.

They work relatively to system/user conf directories and an optional conf
directory given by the environment variable ``B3J0F_CONF_DIR``.
"""

from os import environ, getenv
from os.path import exists, join, expanduser, sep, abspath

from ..base import ConfDriver

from sys import prefix


CONF_DIRS = []  #: all config directories


def _addconfig(config, *paths):
    """Add path to CONF_DIRS if exists."""

    for path in paths:
        if path is not None and exists(path):
            config.append(path)

# add installation directory
_addconfig(CONF_DIRS, join(sep, *(__file__.split(sep)[:-3] + ['data'])))

# add unix system conf
_addconfig(CONF_DIRS, join(sep, 'etc'), join(sep, 'usr', 'local', 'etc'))

_addconfig(
    CONF_DIRS,
    join(prefix, '.config'), join(prefix, 'config'), join(prefix, 'etc')
)

# add XDG conf dirs
_addconfig(CONF_DIRS, getenv('XDG_CONFIG_HOME'), getenv('XDG_CONFIG_DIRS'))

# add windows conf dirs
_addconfig(
    CONF_DIRS,
    getenv('APPDATA'), getenv('PROGRAMDATA'), getenv('LOCALAPPDATA')
)

HOME = environ.get(
    'HOME',  # unix-like
    environ.get(
        'USERPROFILE',  # windows-like
        expanduser('~')
    )  # default
)  #: set user home directory

# add usr config
_addconfig(
    CONF_DIRS, join(HOME, '.config'), join(HOME, 'config'), join(HOME, 'etc')
)

_addconfig(CONF_DIRS, '.')  # add current directory

B3J0F_CONF_DIR = 'B3J0F_CONF_DIR'  #: conf dir environment variable name.

if B3J0F_CONF_DIR in environ:  # add b3j0F_CONF_DIR if necessary
    CONF_DIRS.append(environ[B3J0F_CONF_DIR])  #: conf dir environment variable


class FileConfDriver(ConfDriver):
    """Conf Manager dedicated to files."""

    def rscpaths(self, path):

        result = list(
            join(conf_dir, path) for conf_dir in CONF_DIRS
            if exists(join(conf_dir, path))
        )

        rel_path = expanduser(path)  # add relative path
        if exists(rel_path):
            result.append(rel_path)

        else:
            abs_path = abspath(path)  # add absolute path
            if exists(abs_path):
                result.append(abs_path)

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""model.base UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..base import ModelElement, CompositeModelElement


class ModelElementTest(UTCase):
    """Test ModelElement."""

    class TestME(ModelElement):
        """ModelElement test."""

        __slots__ = ModelElement.__slots__ + ('name', 'cleaned', 'local')

        def __init__(
                self, name=None, local=False, cleaned=False, *args, **kwargs
        ):

            super(ModelElementTest.TestME, self).__init__(
                *args, **kwargs
            )

            self.cleaned = cleaned
            self.name = name
            self.local = local

    def setUp(self):

        self.me = ModelElementTest.TestME()

    def test_copy(self):
        """Test to copy a model element."""

        me = self.me.copy()

        self.assertFalse(me.cleaned)

    def test_cleaned_copy(self):
        """Test to a copied model element."""

        me = self.me.copy(cleaned=True)

        self.assertTrue(me.cleaned)

    def test_update(self):
        """Test the method update."""

        self.assertIsNone(self.me.name)

        copiedme = self.me.copy()

        self.assertIsNone(copiedme.name)

        copiedme.name = 'test'

        self.me.update(copiedme)

        self.assertEqual(self.me.name, 'test')

        self.assertFalse(self.me.cleaned)

        self.me.update(copiedme, copy=True, cleaned=True)

        self.assertTrue(self.me.cleaned)

        copiedme.name = None

        self.me.update(copiedme)

        self.assertIsNone(copiedme.name)
        self.assertIsNotNone(self.me.name)

    def test_update_inheritance(self):
        """Test the method update with inheritance of properties."""

        name = ModelElementTest.TestME(name=1, local=None)
        local = ModelElementTest.TestME(local=2)

        name.update(local)

        self.assertEqual(name.name, 1)
        self.assertEqual(name.local, 2)


class CompositeModelElementTest(UTCase):
    """Test the CompositeModelElement."""

    class TestCME(CompositeModelElement):
        """CompositeModelElement test."""

        __contenttype__ = ModelElementTest.TestME

    def setUp(self):

        self.count = 50

        content = [
            ModelElementTest.TestME(name=str(i)) for i in range(self.count)
        ]

        self.cme = CompositeModelElementTest.TestCME(melts=content)

    def test___iter__(self):
        """Test the __iter__ method."""

        iterator = iter(self.cme.values())

        counts = list(range(self.count))

        for i in range(self.count):

            item = next(iterator)
            counts.remove(int(item.name))

        self.assertEqual(len(counts), 0)

    def test___getitem__(self):
        """Test the __getitem__ method."""

        mes = list(self.cme.values())

        for i in range(self.count):

            me = self.cme[str(i)]
            me = self.cme[me.name]
            mes.remove(me)

        self.assertEqual(0, len(mes))

    def test___setitem__(self):
        """Test the __setitem__ method."""

        for i in range(self.count):

            me = ModelElement()
            self.cme[str(i)] = me

        for me in self.cme:

            self.assertFalse(hasattr(me, 'name'))

        for i in range(self.count):

            me = ModelElementTest.TestME(name=str(i))
            self.cme[me.name] = me

        names = set(me.name for me in self.cme.values())

        self.assertEqual(len(names), self.count)

    def test___delitem__(self):
        """Test the __delitem__ method."""

        for i in range(self.count):

            del self.cme[str(i)]

        self.assertEqual(len(self.cme), 0)

        for i in range(self.count):

            me = ModelElementTest.TestME(name=str(i))
            self.cme[str(i)] = me
            del self.cme[me.name]

        self.assertEqual(len(self.cme), 0)

    def test___getattr__(self):
        """Test the method __getattr__."""

        for i in range(self.count):

            me = getattr(self.cme, str(i))
            self.assertEqual(me.name, str(i))

    def test___contains__(self):
        """Test the method __contains__."""

        for i in range(self.count):

            me = self.cme[str(i)]

            self.assertIn(str(i), self.cme)
            self.assertIn(me.name, self.cme)

        self.assertNotIn(str(i + 1), self.cme)

    def test___len__(self):
        """Test the method __len__."""

        self.assertEqual(len(self.cme), self.count)

        self.cme.clear()

        self.assertEqual(len(self.cme), 0)

    def test___iadd__(self):
        """Test the method __iadd__."""

        cmelen = len(self.cme)

        for i in range(self.count):

            me = self.cme.pop(str(i))

            _cmelen = len(self.cme)
            self.assertEqual(_cmelen, cmelen - 1)

            self.cme += me

            _cmelen = len(self.cme)
            self.assertEqual(_cmelen, cmelen)
            self.assertFalse(self.cme[str(i)].cleaned)

            me = me.copy()
            me.cleaned = True

            self.cme += me

            self.assertEqual(_cmelen, len(self.cme))
            self.assertTrue(self.cme[str(i)].cleaned)

    def test___iadd__me(self):
        """Test the method __iadd__ with a CompositeModelElement."""

        ccme = self.cme.copy()

        ccme.clear()

        self.assertFalse(ccme)

        ccme += self.cme.values()

        self.assertEqual(len(ccme), len(self.cme))

    def test___iadd__iterable(self):
        """Test the method __iadd__ with an iterable."""

        ccme = self.cme.copy()

        ccme.clear()

        self.assertFalse(ccme)

        ccme += list(self.cme.values())

        self.assertEqual(len(ccme), len(self.cme))

    def test___iadd__error(self):
        """Test the method __iadd__ without ModelElement."""

        self.assertRaises(TypeError, self.cme.__iadd__, None)

    def test___isub__me(self):
        """Test the method __isub__ with a model element."""

        for i in range(self.count):

            self.cme -= self.cme[str(i)]

        self.assertFalse(self.cme)

    def test___isub__mes(self):
        """Test the method __isub__ with model elements."""

        self.cme -= list(self.cme.values())

        self.assertFalse(self.cme)

    def test___isub__cme(self):
        """Test the method __isub__ with a CompositeModelElement."""

        self.cme -= self.cme.values()

        self.assertFalse(self.cme)

    def test___isub__error(self):
        """Test the method __isub__ without ModelElement."""

        self.assertRaises(TypeError, self.cme.__isub__, None)

    def test___ixor__(self):
        """Test the method __ixor__."""

        cmelen = len(self.cme)

        for i in range(self.count):

            me = self.cme.pop(str(i))

            _cmelen = len(self.cme)
            self.assertEqual(_cmelen, cmelen - 1)

            self.cme ^= me

            _cmelen = len(self.cme)
            self.assertEqual(_cmelen, cmelen)
            self.assertFalse(self.cme[str(i)].cleaned)

            me = me.copy()
            me.cleaned = True

            self.cme ^= me

            self.assertEqual(_cmelen, len(self.cme))
            self.assertFalse(self.cme[str(i)].cleaned)

    def test___ixor__me(self):
        """Test the method __ixor__ with a CompositeModelElement."""

        ccme = self.cme.copy()

        ccme.clear()

        self.assertFalse(ccme)

        ccme ^= self.cme.values()

        self.assertEqual(len(ccme), len(self.cme))

    def test___ixor__iterable(self):
        """Test the method __ixor__ with an iterable."""

        ccme = self.cme.copy()

        ccme.clear()

        self.assertFalse(ccme)

        ccme ^= list(self.cme.values())

        self.assertEqual(len(ccme), len(self.cme))

    def test___ixor__error(self):
        """Test the method __ixor__ without ModelElement."""

        self.assertRaises(TypeError, self.cme.__ixor__, None)

    def test_keys(self):
        """Test the method keys."""

        keys = tuple(self.cme.keys())
        names = tuple(str(i) for i in range(self.count))

        self.assertEqual(names, keys)

    def test_copy(self):
        """Test the method copy."""

        ccme = self.cme.copy()

        for melt in ccme.values():

            selfmelt = self.cme[melt.name]

            self.assertIsNot(selfmelt, melt)
            self.assertFalse(melt.cleaned)

    def test_pop(self):
        """Test to pop elements."""

        name = str(0)

        self.assertIn(name, self.cme)

        me = self.cme.pop(name)

        self.assertNotIn(name, self.cme)

        me.name == name

    def test_get(self):
        """Test the method get."""

        for elt in list(self.cme.values()):

            melt = self.cme.get(elt.name)

            self.assertIs(elt, melt)

            melt = self.cme.get('')
            self.assertIsNone(melt)

            melt = self.cme.get('', 2)
            self.assertIs(melt, 2)

    def test_setdefault(self):
        """Test the method setdefault."""

        for elt in list(self.cme.values()):

            melt = elt.copy(cleaned=True)
            self.cme.setdefault(melt.name, melt)

            elt = self.cme[melt.name]

            self.assertFalse(elt.cleaned)

    def test_clear(self):
        """Test the method clear."""

        self.assertEqual(len(self.cme), self.count)

        self.cme.clear()

        self.assertFalse(self.cme)

    def test_update(self):
        """Test the method update."""

        me = ModelElementTest.TestME(name='test')
        cme = CompositeModelElementTest.TestCME(melts=[me])

        self.cme.update(cme, copy=True)

        for selfme in self.cme.values():

            if me.name == 'test':
                self.assertFalse(me.local)

            else:
                self.assertTrue(me.local)

    def test_update_inheritance(self):
        """Test the method update with inheritance of properties."""

        name = ModelElementTest.TestME(
            name='test', local=None, cleaned='cleaned'
        )
        local = ModelElementTest.TestME(
            name='test', local='local', cleaned=None
        )

        cname = CompositeModelElementTest.TestCME(melts=[name])
        clocal = CompositeModelElementTest.TestCME(melts=[local])

        lcme = cname.copy()

        lcme.update(clocal)

        self.assertEqual(lcme['test'].local, 'local')
        self.assertEqual(lcme['test'].cleaned, 'cleaned')

        params = lcme.params

        self.assertEqual(params['test'].local, 'local')
        self.assertEqual(params['test'].cleaned, 'cleaned')

if __name__ == '__main__':
    main()



"""Package b3j0f.model."""

__all__ = [
    'Configuration', 'configuration',
    'Category', 'category',
    'Parameter', 'Array', 'BOOL', 'ARRAY', 'PType'
]

from .conf import Configuration, configuration
from .cat import Category, category
from .param import Parameter, BOOL, ARRAY, Array, PType



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""JSON configuration file driver."""

from __future__ import absolute_import

__all__ = ['JSONFileConfDriver']

try:
    from json import load, dump

except ImportError:
    from simplejson import load, dump

from .base import FileConfDriver
from ..json import JSONConfDriver


class JSONFileConfDriver(FileConfDriver, JSONConfDriver):
    """Manage json resource configuration from json file."""

    def _pathresource(self, rscpath):

        result = None

        with open(rscpath, 'r') as fpr:

            result = load(fpr)

        return result

    def _setconf(self, conf, resource, rscpath):

        super(JSONFileConfDriver, self)._setconf(conf, resource, rscpath)

        with open(rscpath, 'w') as fpw:

            dump(resource, fpw)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""python expression resolver."""

__all__ = ['resolvepy']

from b3j0f.utils.runtime import safe_eval
from b3j0f.utils.path import lookup

from re import compile as re_compile, sub

from ..registry import register
from ..core import (
    DEFAULT_BESTEFFORT, DEFAULT_SAFE, DEFAULT_TOSTR, DEFAULT_SCOPE
)


MISSING_NAME = r'\'(?P<name>\w+)\''
MISSING_NAME_REGEX = re_compile(MISSING_NAME)

MISSING_VARIABLE = r'(?P<name>(\w+\.)*{0}(\.\w+)*)'


def genrepl(scope):
    """Replacement function with specific scope."""

    def repl(match):
        """Internal replacement function."""

        name = match.group('name')

        value = lookup(name, scope=scope)

        result = name.replace('.', '_')
        scope[result] = value

        return result

    return repl


@register('py')
@register('python')
def resolvepy(
        expr,
        safe=DEFAULT_SAFE, tostr=DEFAULT_TOSTR, scope=DEFAULT_SCOPE,
        besteffort=DEFAULT_BESTEFFORT
    ):
    """Resolve input expression.

    :param str expr: configuration expression to resolve in this language.
    :param bool safe: safe run execution context (True by default).
    :param bool tostr: format the result.
    :param dict scope: execution scope (contains references to expression
        objects).
    :param bool besteffort: try to resolve unknown variable name with execution
        runtime."""

    result = None

    _eval = safe_eval if safe else eval

    _expr = expr

    scope = {} if scope is None else scope.copy()

    while True:

        try:
            result = _eval(_expr, scope)

        except (AttributeError, NameError) as nex:

            if not besteffort:
                raise

            arg = nex.args[0]
            missing = MISSING_NAME_REGEX.findall(arg)[-1]

            try:
                _expr = sub(
                    MISSING_VARIABLE.format(missing),
                    genrepl(scope=scope),
                    _expr
                )

            except ImportError:
                raise nex

        else:
            break

    if tostr:
        result = str(result)

    return result
